#include <string.h>

#include "winternitz.h"
#include "randombytes.h"
#include "shake256.h"

int crypto_ots_keypair(unsigned char *pk, unsigned char *sk) 
{
  int i,j;
  randombytes(sk,CRYPTO_OTS_SECRETKEYBYTES);
  
  for(i=0;i<WOTS_L;i++)
  {
    memcpy(pk,sk,CRYPTO_OTS_HASHBYTES);
    for(j=0;j<WOTS_W-1;j++)
      shake256(pk,CRYPTO_OTS_HASHBYTES,pk,CRYPTO_OTS_HASHBYTES);
    pk += CRYPTO_OTS_HASHBYTES;
    sk += CRYPTO_OTS_HASHBYTES;
  }
  return 0;
}

#if (WOTS_LOGW == 4)
int crypto_ots_sign(unsigned char *sig,
                    const unsigned char *m,
                    unsigned long long mlen, 
                    const unsigned char *sk) 
{
  unsigned char h[CRYPTO_OTS_HASHBYTES];
  int i,j,t;
  unsigned int checksum=WOTS_L*(WOTS_W-1);


  shake256(h,CRYPTO_OTS_HASHBYTES,m,mlen);

  for(i=0;i<WOTS_L1;i+=2)
  {
    memcpy(sig, sk, CRYPTO_OTS_HASHBYTES);
    t = h[i/2] & 0xf;
    for(j=0;j<t;j++)
      shake256(sig,CRYPTO_OTS_HASHBYTES,sig,CRYPTO_OTS_HASHBYTES);
    checksum -= t;
    sig += CRYPTO_OTS_HASHBYTES;
    sk  += CRYPTO_OTS_HASHBYTES;

    memcpy(sig, sk, CRYPTO_OTS_HASHBYTES);
    t = h[i/2] >> 4;
    for(j=0;j<t;j++)
      shake256(sig,CRYPTO_OTS_HASHBYTES,sig,CRYPTO_OTS_HASHBYTES);
    checksum -= t;
    sig += CRYPTO_OTS_HASHBYTES;
    sk  += CRYPTO_OTS_HASHBYTES;
  }
  
  for(i=0;i<WOTS_L2;i++)
  {
    memcpy(sig, sk, CRYPTO_OTS_HASHBYTES);
    t = checksum & 0xf;
    for(j=0;j<t;j++)
      shake256(sig,CRYPTO_OTS_HASHBYTES,sig,CRYPTO_OTS_HASHBYTES);
    checksum >>= 4;
    sig += CRYPTO_OTS_HASHBYTES;
    sk  += CRYPTO_OTS_HASHBYTES;
  }
  return 0;
}
#else
#error "crypto_ots_sign assumes WOTS_LOGW == 4."
#endif

#if (WOTS_LOGW == 4)
int crypto_ots_verify(unsigned char *pk,
                      const unsigned char *m,
                      unsigned long long mlen,
                      const unsigned char *sig)
{
  unsigned char h[CRYPTO_OTS_HASHBYTES];
  int i,j,t;
  unsigned int checksum=WOTS_L*(WOTS_W-1);


  shake256(h,CRYPTO_OTS_HASHBYTES,m,mlen);

  for(i=0;i<WOTS_L1;i+=2)
  {
    memcpy(pk, sig, CRYPTO_OTS_HASHBYTES);
    t = h[i/2] & 0xf;
    for(j=t;j<WOTS_W-1;j++)
      shake256(pk,CRYPTO_OTS_HASHBYTES,pk,CRYPTO_OTS_HASHBYTES);
    checksum -= t;
    pk  += CRYPTO_OTS_HASHBYTES;
    sig += CRYPTO_OTS_HASHBYTES;

    memcpy(pk, sig, CRYPTO_OTS_HASHBYTES);
    t = h[i/2] >> 4;
    for(j=t;j<WOTS_W-1;j++)
      shake256(pk,CRYPTO_OTS_HASHBYTES,pk,CRYPTO_OTS_HASHBYTES);
    checksum -= t;
    pk  += CRYPTO_OTS_HASHBYTES;
    sig += CRYPTO_OTS_HASHBYTES;
  }
  
  for(i=0;i<WOTS_L2;i++)
  {
    memcpy(pk, sig, CRYPTO_OTS_HASHBYTES);
    t = checksum & 0xf;
    for(j=t;j<WOTS_W-1;j++)
      shake256(pk,CRYPTO_OTS_HASHBYTES,pk,CRYPTO_OTS_HASHBYTES);
    checksum >>= 4;
    pk  += CRYPTO_OTS_HASHBYTES;
    sig += CRYPTO_OTS_HASHBYTES;
  }
  return 0;
}
#else
#error "crypto_ots_verify assumes WOTS_LOGW == 4."
#endif
